/**
 * ## When should I use this?
 *
 * You can use this package to check that referenced definitions are defined.
 *
 * ## API
 *
 * The following options (default: `undefined`) are accepted:
 *
 * *   `Object` with the following fields:
 *     *   `allow` (`Array<string | RegExp | { source: string }>`,
 *         default: `[]`)
 *         — text or regex that you want to be allowed between `[` and `]`
 *         even though it’s undefined; regex is provided via a `RegExp` object
 *         or via a `{ source: string }` object where `source` is the source
 *         text of a case-insensitive regex
 *
 * ## Recommendation
 *
 * Shortcut references use an implicit syntax that could also occur as plain
 * text.
 * For example, it is reasonable to expect an author adding `[…]` to abbreviate
 * some text somewhere in a document:
 *
 * ```markdown
 * > Some […] quote.
 * ```
 *
 * This isn’t a problem, but it might become one when an author later adds a
 * definition:
 *
 * ```markdown
 * Some text. […][]
 *
 * […] #read-more "Read more"
 * ```
 *
 * The second author might expect only their newly added text to form a link,
 * but their changes also result in a link for the first author’s text.
 *
 * @module no-undefined-references
 * @summary
 *   remark-lint rule to warn when undefined definitions are referenced.
 * @author Titus Wormer
 * @copyright 2016 Titus Wormer
 * @license MIT
 * @example
 *   {"name": "ok.md"}
 *
 *   [foo][]
 *
 *   Just a [ bracket.
 *
 *   Typically, you’d want to use escapes (with a backslash: \\) to escape what
 *   could turn into a \[reference otherwise].
 *
 *   Just two braces can’t link: [].
 *
 *   [foo]: https://example.com
 *
 * @example
 *   {"name": "ok-allow.md", "config": {"allow": ["...", "…"]}}
 *
 *   > Eliding a portion of a quoted passage […] is acceptable.
 *
 * @example
 *   {"name": "ok-allow.md", "config": {"allow": ["a", {"source": "^b\\."}]}}
 *
 *   [foo][b.c]
 *
 *   [bar][a]
 *
 *   Matching is case-insensitive: [bar][B.C]
 *
 * @example
 *   {"name": "not-ok.md", "label": "input"}
 *
 *   [bar]
 *
 *   [baz][]
 *
 *   [text][qux]
 *
 *   Spread [over
 *   lines][]
 *
 *   > in [a
 *   > block quote][]
 *
 *   [asd][a
 *
 *   Can include [*emphasis*].
 *
 *   Multiple pairs: [a][b][c].
 *
 * @example
 *   {"name": "not-ok.md", "label": "output"}
 *
 *   1:1-1:6: Found reference to undefined definition
 *   3:1-3:8: Found reference to undefined definition
 *   5:1-5:12: Found reference to undefined definition
 *   7:8-8:9: Found reference to undefined definition
 *   10:6-11:17: Found reference to undefined definition
 *   13:1-13:6: Found reference to undefined definition
 *   15:13-15:25: Found reference to undefined definition
 *   17:17-17:23: Found reference to undefined definition
 *   17:23-17:26: Found reference to undefined definition
 *
 * @example
 *   {"name": "not-ok.md", "label": "input", "config": {"allow": ["a", {"source": "^b\\."}]}}
 *
 *   [foo][a.c]
 *
 *   [bar][b]
 *
 * @example
 *   {"name": "not-ok.md", "label": "output", "config": {"allow": ["a", {"source": "^b\\."}]}}
 *
 *   1:1-1:11: Found reference to undefined definition
 *   3:1-3:9: Found reference to undefined definition
 */

/**
 * @typedef {import('mdast').Root} Root
 * @typedef {import('mdast').Heading} Heading
 * @typedef {import('mdast').Paragraph} Paragraph
 *
 * @typedef Options
 * @property {Array<string | RegExp | { source: string }>} [allow]
 *
 * @typedef {Array<number>} Range
 */

import {normalizeIdentifier} from 'micromark-util-normalize-identifier'
import {location} from 'vfile-location'
import {lintRule} from 'unified-lint-rule'
import {generated} from 'unist-util-generated'
import {pointStart, pointEnd} from 'unist-util-position'
import {visit, SKIP, EXIT} from 'unist-util-visit'

const remarkLintNoUndefinedReferences = lintRule(
  {
    origin: 'remark-lint:no-undefined-references',
    url: 'https://github.com/remarkjs/remark-lint/tree/main/packages/remark-lint-no-undefined-references#readme'
  },
  /** @type {import('unified-lint-rule').Rule<Root, Options>} */
  (tree, file, option = {}) => {
    const contents = String(file)
    const loc = location(file)
    const lineEnding = /(\r?\n|\r)[\t ]*(>[\t ]*)*/g
    /** @type {Record<string, boolean>} */
    const map = Object.create(null)

    const allow = option.allow || []
    /** @type {Array<RegExp>} */
    const regexes = []
    /** @type {Set<string>} */
    const strings = new Set()

    let index = -1

    while (++index < allow.length) {
      const value = allow[index]
      if (typeof value === 'string') {
        strings.add(normalizeIdentifier(value))
      } else if (value instanceof RegExp) {
        regexes.push(value)
      } else {
        regexes.push(new RegExp(value.source, 'i'))
      }
    }

    visit(tree, (node) => {
      if (
        (node.type === 'definition' || node.type === 'footnoteDefinition') &&
        !generated(node)
      ) {
        map[normalizeIdentifier(node.identifier)] = true
      }
    })

    visit(tree, (node) => {
      // CM specifiers that references only form when defined.
      // Still, they could be added by plugins, so let’s keep it.
      /* c8 ignore next 10 */
      if (
        (node.type === 'imageReference' ||
          node.type === 'linkReference' ||
          node.type === 'footnoteReference') &&
        !generated(node) &&
        !(normalizeIdentifier(node.identifier) in map) &&
        !isAllowed(node.identifier)
      ) {
        file.message('Found reference to undefined definition', node)
      }

      if (node.type === 'paragraph' || node.type === 'heading') {
        findInPhrasing(node)
      }
    })

    /**
     * @param {Heading|Paragraph} node
     */
    function findInPhrasing(node) {
      /** @type {Array<Range>} */
      let ranges = []

      visit(node, (child) => {
        // Ignore the node itself.
        if (child === node) return

        // Can’t have links in links, so reset ranges.
        if (child.type === 'link' || child.type === 'linkReference') {
          ranges = []
          return SKIP
        }

        // Enter non-text.
        if (child.type !== 'text') return

        const start = pointStart(child).offset
        const end = pointEnd(child).offset

        // Bail if there’s no positional info.
        if (typeof start !== 'number' || typeof end !== 'number') {
          return EXIT
        }

        const source = contents.slice(start, end)
        /** @type {Array<[number, string]>} */
        const lines = [[start, '']]
        let last = 0

        lineEnding.lastIndex = 0
        let match = lineEnding.exec(source)

        while (match) {
          const index = match.index
          lines[lines.length - 1][1] = source.slice(last, index)
          last = index + match[0].length
          lines.push([start + last, ''])
          match = lineEnding.exec(source)
        }

        lines[lines.length - 1][1] = source.slice(last)
        let lineIndex = -1

        while (++lineIndex < lines.length) {
          const line = lines[lineIndex][1]
          let index = 0

          while (index < line.length) {
            const code = line.charCodeAt(index)

            // Skip past escaped brackets.
            if (code === 92) {
              const next = line.charCodeAt(index + 1)
              index++

              if (next === 91 || next === 93) {
                index++
              }
            }
            // Opening bracket.
            else if (code === 91) {
              ranges.push([lines[lineIndex][0] + index])
              index++
            }
            // Close bracket.
            else if (code === 93) {
              // No opening.
              if (ranges.length === 0) {
                index++
              } else if (line.charCodeAt(index + 1) === 91) {
                index++

                // Collapsed or full.
                let range = ranges.pop()

                // Range should always exist.
                // eslint-disable-next-line max-depth
                if (range) {
                  range.push(lines[lineIndex][0] + index)

                  // This is the end of a reference already.
                  // eslint-disable-next-line max-depth
                  if (range.length === 4) {
                    handleRange(range)
                    range = []
                  }

                  range.push(lines[lineIndex][0] + index)
                  ranges.push(range)
                  index++
                }
              } else {
                index++

                // Shortcut or typical end of a reference.
                const range = ranges.pop()

                // Range should always exist.
                // eslint-disable-next-line max-depth
                if (range) {
                  range.push(lines[lineIndex][0] + index)
                  handleRange(range)
                }
              }
            }
            // Anything else.
            else {
              index++
            }
          }
        }
      })

      let index = -1

      while (++index < ranges.length) {
        handleRange(ranges[index])
      }

      return SKIP

      /**
       * @param {Range} range
       */
      function handleRange(range) {
        if (range.length === 1) return
        if (range.length === 3) range.length = 2

        // No need to warn for just `[]`.
        if (range.length === 2 && range[0] + 2 === range[1]) return

        const offset = range.length === 4 && range[2] + 2 !== range[3] ? 2 : 0
        const id = contents
          .slice(range[0 + offset] + 1, range[1 + offset] - 1)
          .replace(lineEnding, ' ')
        const pos = {
          start: loc.toPoint(range[0]),
          end: loc.toPoint(range[range.length - 1])
        }

        if (
          !generated({position: pos}) &&
          !(normalizeIdentifier(id) in map) &&
          !isAllowed(id)
        ) {
          file.message('Found reference to undefined definition', pos)
        }
      }
    }

    /**
     * @param {string} id
     * @returns {boolean}
     */
    function isAllowed(id) {
      const normalized = normalizeIdentifier(id)
      return (
        strings.has(normalized) ||
        regexes.some((regex) => regex.test(normalized))
      )
    }
  }
)

export default remarkLintNoUndefinedReferences
